﻿using System;
using System.Device.Location;
using System.IO.IsolatedStorage;
using System.Text;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Media;
using System.Windows.Media.Imaging;
using System.Windows.Navigation;
using Microsoft.Devices.Sensors;
using Microsoft.Phone.Controls;
using Microsoft.Phone.Shell;
using Microsoft.Xna.Framework;
using RotatingMapTiles.ImageryService;

namespace RotatingMapTiles
{
    public partial class MainPage : PhoneApplicationPage
    {
        const int BITRES = 29;

        GeoCoordinateWatcher geoWatcher = new GeoCoordinateWatcher(GeoPositionAccuracy.High);
        Motion motion;
        StringBuilder strBuilder = new StringBuilder();

        string roadUriTemplate, aerialUriTemplate;
        MapStyle mapStyle;
        int zoomLevel;
        int minimumLevel, maximumLevel;
        int integerLongitude = -1;
        int integerLatitude = -1;

        public MainPage()
        {
            InitializeComponent();

            // Fix that annoying ApplicationBar bug
            minusAppBarIconButton = this.ApplicationBar.Buttons[1] as ApplicationBarIconButton;
            plusAppBarIconButton = this.ApplicationBar.Buttons[2] as ApplicationBarIconButton;

            // Create 25 Image elements for a 5x5 array
            for (int i = 0; i < 25; i++)
            {
                Image image = new Image 
                { 
                    Source = new BitmapImage(),
                    Stretch = Stretch.None 
                };
                imageCanvas.Children.Add(image);
            }

            // Initialize the Motion sensor
            if (Motion.IsSupported)
            {
                motion = new Motion();
                motion.TimeBetweenUpdates = TimeSpan.FromMilliseconds(33);
                motion.CurrentValueChanged += OnMotionCurrentValueChanged;
            }

            // Two more event handlers
            geoWatcher.PositionChanged += OnGeoWatcherPositionChanged;
            Loaded += OnMainPageLoaded;
        }

        protected override void OnNavigatedTo(NavigationEventArgs args)
        {
            // Get saved information
            IsolatedStorageSettings settings = IsolatedStorageSettings.ApplicationSettings;

            if (!settings.TryGetValue<int>("ZoomLevel", out zoomLevel))
                zoomLevel = 12;

            if (!settings.TryGetValue<MapStyle>("MapStyle", out mapStyle))
                mapStyle = MapStyle.Road;

            // Start the Motion sensor GPS watcher
            if (motion != null)
            {
                try { motion.Start(); }
                catch { }
            }

            geoWatcher.Start();
            base.OnNavigatedTo(args);
        }

        protected override void OnNavigatedFrom(NavigationEventArgs args)
        {
            // Stop the Motion sensor and GPS watcher
            if (motion != null)
            {
                motion.Stop();
            }

            geoWatcher.Stop();

            // Save information
            IsolatedStorageSettings settings = IsolatedStorageSettings.ApplicationSettings;
            settings["ZoomLevel"] = zoomLevel;
            settings["MapStyle"] = mapStyle;
            settings.Save();

            base.OnNavigatedFrom(args);
        }

        void OnMainPageLoaded(object sender, RoutedEventArgs e)
        {
            // Initialize the Bing Maps imagery service
            ImageryServiceClient imageryServiceClient = new ImageryServiceClient("BasicHttpBinding_IImageryService");
            imageryServiceClient.GetImageryMetadataCompleted += GetImageryMetadataCompleted;

            // Make a request for the road metadata
            ImageryMetadataRequest request = new ImageryMetadataRequest
            {
                Credentials = new Credentials
                {
                    ApplicationId = "credentials-key"
                },
                Style = MapStyle.Road
            };
            imageryServiceClient.GetImageryMetadataAsync(request, "road");

            // Make a request for the aerial metadata
            request.Style = MapStyle.Aerial;
            imageryServiceClient.GetImageryMetadataAsync(request, "aerial");
        }

        void GetImageryMetadataCompleted(object sender, GetImageryMetadataCompletedEventArgs args)
        {
            if (!args.Cancelled && args.Error == null)
            {
                // Get the "powered by" bitmap
                poweredByBitmap.UriSource = args.Result.BrandLogoUri;
                poweredByDisplay.Visibility = Visibility.Visible;

                // Get the range of map levels available
                ImageryMetadataResult result = args.Result.Results[0];
                minimumLevel = result.ZoomRange.From;
                maximumLevel = result.ZoomRange.To;

                // Get the URI and make some substitutions
                string uri = result.ImageUri;
                uri = uri.Replace("{subdomain}", result.ImageUriSubdomains[0]);
                uri = uri.Replace("&token={token}", "");
                uri = uri.Replace("{culture}", "en-US");

                if (args.UserState as string == "road")
                    roadUriTemplate = uri;
                else
                    aerialUriTemplate = uri;

                if (roadUriTemplate != null && aerialUriTemplate != null)
                    RefreshDisplay();
            }
            else
            {
                errorTextBlock.Text = "Cannot access Bing Maps: " + args.Error.Message;
            }
        }

        void OnMotionCurrentValueChanged(object sender, SensorReadingEventArgs<MotionReading> args)
        {
            this.Dispatcher.BeginInvoke(() =>
            {
                // Get the rotation matrix & convert to horizontal coordinates
                Microsoft.Xna.Framework.Matrix matrix = args.SensorReading.Attitude.RotationMatrix;
                Vector3 vector = Vector3.Transform(Vector3.UnitY, matrix);

                // Set the azimuth angle to rotation transforms
                double azimuth = 180 * Math.Atan2(vector.X, vector.Y) / Math.PI;
                imageCanvasRotate.Angle = azimuth;
                northArrowRotate.Angle = azimuth;
            });
        }

        void OnGeoWatcherPositionChanged(object sender,
                                 GeoPositionChangedEventArgs<GeoCoordinate> args)
        {
            // Make the location cirlc visible
            locationDisplay.Visibility = Visibility.Visible;

            // Save the integer longitude and latitude
            double longitude = args.Position.Location.Longitude;
            double latitude = args.Position.Location.Latitude;

            double relativeLongitude = (180 + longitude) / 360;
            integerLongitude = (int)(relativeLongitude * (1 << BITRES));

            double sinTerm = Math.Sin(Math.PI * latitude / 180);
            double relativeLatitude = 0.5 - Math.Log((1 + sinTerm) / (1 - sinTerm)) / (4 * Math.PI);
            integerLatitude = (int)(relativeLatitude * (1 << BITRES));

            RefreshDisplay();
        }

        void OnToggleViewAppBarButtonClick(object sender, EventArgs args)
        {
            if (mapStyle == MapStyle.Road)
                mapStyle = MapStyle.Aerial;
            else
                mapStyle = MapStyle.Road;

            RefreshDisplay();
        }

        void OnZoomInAppBarButtonClick(object sender, EventArgs args)
        {
            zoomLevel += 1;
            EnableZoomButtons();
            RefreshDisplay();
        }

        void OnZoomOutAppBarButtonClick(object sender, EventArgs args)
        {
            zoomLevel -= 1;
            EnableZoomButtons();
            RefreshDisplay();
        }

        void EnableZoomButtons()
        {
            minusAppBarIconButton.IsEnabled = zoomLevel > minimumLevel;
            plusAppBarIconButton.IsEnabled = zoomLevel < maximumLevel;
        }

        void RefreshDisplay()
        {
            if (roadUriTemplate == null || aerialUriTemplate == null)
                return;

            if (integerLongitude == -1 || integerLatitude == -1)
                return;

            // Get coordinates and pixel offsets based on current zoom level
            int croppedLongitude = integerLongitude >> BITRES - zoomLevel;
            int croppedLatitude = integerLatitude >> BITRES - zoomLevel;
            int xPixelOffset = (integerLongitude >> BITRES - zoomLevel - 8) % 256;
            int yPixelOffset = (integerLatitude >> BITRES - zoomLevel - 8) % 256;

            // Prepare for the loop
            string uriTemplate = mapStyle == MapStyle.Road ? roadUriTemplate : aerialUriTemplate;
            int index = 0;
            int maxValue = (1 << zoomLevel) - 1;

            // Loop through the 5x5 array of Image elements
            for (int row = -2; row <= 2; row++)
                for (int col = -2; col <= 2; col++)
                {
                    // Get the Image and BitmapImage
                    Image image = imageCanvas.Children[index] as Image;
                    BitmapImage bitmap = image.Source as BitmapImage;
                    index++;

                    // Check if we've gone beyond the bounds
                    if (croppedLongitude + col < 0 ||
                        croppedLongitude + col > maxValue ||
                        croppedLatitude + row < 0 ||
                        croppedLatitude + row > maxValue)
                    {
                        bitmap.UriSource = null;
                    }
                    else
                    {
                        // Calculate a quadkey and set URI to bitmap
                        int longitude = croppedLongitude + col;
                        int latitude = croppedLatitude + row;
                        string strQuadkey = ToQuadKey(longitude, latitude, zoomLevel);
                        string uri = uriTemplate.Replace("{quadkey}", strQuadkey);
                        bitmap.UriSource = new Uri(uri);
                    }

                    // Position the Image element
                    Canvas.SetLeft(image, col * 256 - xPixelOffset);
                    Canvas.SetTop(image, row * 256 - yPixelOffset);
                }
        }

        string ToQuadKey(int longitude, int latitude, int level)
        {
            long quadkey = 0;
            int mask = 1 << (level - 1);

            for (int i = 0; i < level; i++)
            {
                quadkey <<= 2;

                if ((longitude & mask) != 0)
                    quadkey |= 1;

                if ((latitude & mask) != 0)
                    quadkey |= 2;

                mask >>= 1;
            }

            strBuilder.Clear();

            for (int i = 0; i < level; i++)
            {
                strBuilder.Insert(0, (quadkey & 3).ToString());
                quadkey >>= 2;
            }

            return strBuilder.ToString();
        }
    }
}